/*
  Name: VM.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This file contains the Virtual machine which will execute byte code
	  passed to it. It will also return instructions to the game and the map
*/

#include <allegro.h>
#include <almp3.h>
#include <libnet.h>
#include <string>
using namespace std;

#include "Sound.h"
#include "messageClass.h"
#include "bytecode.h"
#include "Compiler.h"
#include "VM.h"
#include "ElementaryFunctions.h"
#include "Errors.h"
#include "Global.h"
#include "keyStatesClass.h"

VirtualMachine::VirtualMachine()
{
	for(int i = 0; i < ADDRESSRANGE; ++i)
		var[i] = 0;
	//Set the TRUE constant to 1
	var[10] = 1;
	needsNextMap = false;
}

void VirtualMachine::compute(bytecode &code)
{
	if(!code.getLength())
		return;
	internalCode = &code;
	code.reset();
	unsigned char buffer;
	int ends = 0;
	bool run = true;
	while(run)
	{
		code.read(&buffer, 1);
		switch(buffer)
		{
		case SCRIPTEND:
			run = false;
			break;
			//Set variable to value
		case SET:
			{
				int dest = getVarPar();
				getVar(dest, true) = getPar();
				//allegro_message("%i", getVar(dest, false));
			}
			break;
			//Increment variable by one
		case INC:
			{
				++getVar(getVarPar(), true);
			}
			break;
			//Decrement variable by one
		case DEC:
			{
				--getVar(getVarPar(), true);
			}
			break;
			//Add value to variable
		case ADD:
			{
				int dest = getVarPar();
				getVar(dest, true) += getPar();
			}
			break;
			//Substract value from variable
		case SUB:
			{
				int dest = getVarPar();
				getVar(dest, true) -= getPar();
			}
			break;
			//Multiplie variable with value
		case MUL:
			{
				int value = getVarPar();
				getVar(value, true) *= getPar();
			}
			break;
		case DIV:
			{
				int value = getVarPar();
				getVar(value, true) /= getPar();
			}
			break;
		case IF:
			{
				if(getPar(true))
					++ends;
				else
				{
					int internalEnds = 1;
					while(true)
					{
						code.read(&buffer, 1);
						if(buffer == IFEND)
						{
							--internalEnds;
							if(!internalEnds)
								break;
						}
						if(buffer == IF || buffer == IFN)
						{
							++internalEnds;
						}
					}
				}
			}
			break;
		case IFN:
			{
				if(!getPar(true))
					++ends;
				else
				{
					int internalEnds = 1;
					while(true)
					{
						code.read(&buffer, 1);
						if(buffer == IFEND)
						{
							--internalEnds;
							if(!internalEnds)
								break;
							continue;
						}
						if(buffer == IF || buffer == IFN)
						{
							++internalEnds;
						}
					}
				}
			}
			break;
		case IFEND:
			{
				--ends;
				if(ends < 0)
					errors::scriptError("Too many ends in script!");
					//errors::scriptError(string("Too many ends in script at position ") + i2str(code.getPos()));
			}
			break;
		case EQUAL:
			{
				//Write to ZFLAG if the two values are equal
				getVar(12) = getPar(true) == getPar(false);
			}
			break;
		case SMALLER:
			{
				int vars[2] = { getPar(), getPar() };
				//Write to ZFLAG if they are smaller
				getVar(12) = (vars[0] < vars[1]);
			}
			break;
		case QUIT:
			{
				if(getPar(true))
					run = false;
			}
			break;
		case QUITN:
			{
				if(!getPar(true))
					run = false;
			}
			break;
		case QUITA:
			{
				run = false;
			}
			break;
		case LOADSOUND:
			{
				int varIndex = getVarPar();
				string temp = code.getString(0x00);
				for(int i = 0; i < MAXSOUNDS; ++i)
				{
					if(!sounds[i].isActive())
					{
						sounds[i].load(temp, true);
						getVar(varIndex, true) = i;
						break;
					}
					if(i + 1 == MAXSOUNDS)
						errors::scriptError("Sound buffer is full!");
				}
			}
			break;
		case DELETESOUND:
			{
				int index = getPar(true);
				if(index >= MAXSOUNDS)
					index = 0;
				sounds[index].destroy();
			}
			break;
		case PLAYSOUND:
			{
				int varIndex = getPar(true);
				int values[2] = { getPar(), getPar() };
				int volume = global.sfxVol;
				//Music file
				if(values[1])
					volume = global.musicVol;
				if(varIndex >= MAXSOUNDS)
					varIndex = 0;
				sounds[varIndex].play(values[0], volume);
			}
			break;
		case STOPSOUND:
			{
				sounds[getPar(true)].stop();
			}
			break;
		case GRAPHIC:
			{
				int values[3] = { getPar(), getPar(), getPar() };
				graphicData[getOffset(values[0], values[1])] = char(values[2]);
			}
			break;
		case GOTO:
			{
				nextMap = code.getString(0x00);
				needsNextMap = true;
			}
			break;
		case MSG:
			{
				messages->sendText(code.getString(0x00));
			}
			break;
		case MSGVAR:
			{
				messages->sendInt(getPar(true));
			}
			break;
		case ENDL:
			{
				messages->output();
			}
			break;
		case MSGKEY:
			{
				int varIndex = getPar();
				if(varIndex > 11)
					errors::scriptError("msgkey value out of range!");
				//A hack to display the name of a key
				messages->sendText(keyNames[(&global.keyLeft)[varIndex]]);
			}
			break;
		default:
			allegro_message("Command %i at position %i, missing %i ends", buffer, code.getPos(), ends);
			errors::scriptError("Unknown command");
		}
	}
	var[8] = true;
	for(int i = 0; i < objectives; ++i)
		if(!var[18 + i])
		{
			var[8] = false;
			break;
		}
}

inline int VirtualMachine::getOffset(int x, int y)
{
	return x + mapWidth * y;
}

void VirtualMachine::mapSetup(int w, int h, char *data)
{
	mapWidth = w;
	mapHeight = h;
	graphicData = data;
}

//inline int &VirtualMachine::getVar(int offset, bool write)

bool VirtualMachine::needNextMap()
{
	return needsNextMap;
}

string VirtualMachine::getMapString()
{
	return nextMap;
}

void VirtualMachine::setMessageClass(messageClass *msc)
{
	messages = msc;
}

void VirtualMachine::setObjectives(int obj)
{
	objectives = obj;
}

void VirtualMachine::pollSounds()
{
	for(int i = 0; i < MAXSOUNDS; ++i)
		sounds[i].poll();
}

void VirtualMachine::updateSoundVol()
{
	for(int i = 0; i < MAXSOUNDS; ++i)
		sounds[i].musicUpdate();
}

int VirtualMachine::getPar(bool varOnly)
{
	if(varOnly)
		return getVar(getVarPar());
	if(internalCode->isVar())
	{
		return getVar(internalCode->getVarPar());
	}
	else
	{
		return internalCode->getConstPar();
	}
}

int VirtualMachine::getVarPar()
{
	if(!internalCode->isVar())
		errors::scriptError("Unexpected constant value!");
	return internalCode->getVarPar();
}
